"""
functions map tensor to scale
"""
import numpy as np


def rmse(label, target):
    err = target - label
    value = np.sqrt(np.sum(np.square(err)) / label.size)
    # base = np.sqrt(np.sum(np.square(target)) / label.size)
    # return value / base
    return value

def mean_absolute_error(label, target):
    err = target - label
    return np.mean(np.abs(err))

def bias(label, target):
    return np.mean(label - target)


def variance(label, target):
    return np.std(label - target)


def psnr(label, target):
    minv = np.min(label)
    maxv = np.max(label)
    sca = 255.0 / (maxv - minv)
    ln = (label - minv) * sca
    tn = (target - minv) * sca
    rmv = rmse(ln, tn)
    value = 10 * np.log((255.0**2) / (rmv**2)) / np.log(10)
    return value


'''
The function to compute SSIM
@param param: img_mat_1 1st 2D matrix
@param param: img_mat_2 2nd 2D matrix
'''


def ssim(label, target):
    minv = np.min(label)
    maxv = np.max(label)
    sca = 255.0 / (maxv - minv)
    img_mat_1 = (label - minv) * sca
    img_mat_2 = (target - minv) * sca
    import numpy
    import scipy.ndimage
    from numpy.ma.core import exp
    from scipy.constants.constants import pi
    # Variables for Gaussian kernel definition
    gaussian_kernel_sigma = 1.5
    gaussian_kernel_width = 11
    gaussian_kernel = numpy.zeros(
        (gaussian_kernel_width, gaussian_kernel_width))

    # Fill Gaussian kernel
    for i in range(gaussian_kernel_width):
        for j in range(gaussian_kernel_width):
            gaussian_kernel[i, j] =\
                (1 / (2 * pi * (gaussian_kernel_sigma**2))) *\
                exp(-(((i - 5)**2) + ((j - 5)**2)) /
                    (2 * (gaussian_kernel_sigma**2)))

    # Convert image matrices to double precision (like in the Matlab version)
    img_mat_1 = img_mat_1.astype(numpy.float)
    img_mat_2 = img_mat_2.astype(numpy.float)

    # Squares of input matrices
    img_mat_1_sq = img_mat_1**2
    img_mat_2_sq = img_mat_2**2
    img_mat_12 = img_mat_1 * img_mat_2

    # Means obtained by Gaussian filtering of inputs
    img_mat_mu_1 = scipy.ndimage.filters.convolve(img_mat_1, gaussian_kernel)
    img_mat_mu_2 = scipy.ndimage.filters.convolve(img_mat_2, gaussian_kernel)

    # Squares of means
    img_mat_mu_1_sq = img_mat_mu_1**2
    img_mat_mu_2_sq = img_mat_mu_2**2
    img_mat_mu_12 = img_mat_mu_1 * img_mat_mu_2

    # Variances obtained by Gaussian filtering of inputs' squares
    img_mat_sigma_1_sq = scipy.ndimage.filters.convolve(
        img_mat_1_sq, gaussian_kernel)
    img_mat_sigma_2_sq = scipy.ndimage.filters.convolve(
        img_mat_2_sq, gaussian_kernel)

    # Covariance
    img_mat_sigma_12 = scipy.ndimage.filters.convolve(
        img_mat_12, gaussian_kernel)

    # Centered squares of variances
    img_mat_sigma_1_sq = img_mat_sigma_1_sq - img_mat_mu_1_sq
    img_mat_sigma_2_sq = img_mat_sigma_2_sq - img_mat_mu_2_sq
    img_mat_sigma_12 = img_mat_sigma_12 - img_mat_mu_12

    # c1/c2 constants
    # First use: manual fitting
    c_1 = 6.5025
    c_2 = 58.5225

    # Second use: change k1,k2 & c1,c2 depend on L (width of color map)
    l = 255
    k_1 = 0.01
    c_1 = (k_1 * l)**2
    k_2 = 0.03
    c_2 = (k_2 * l)**2

    # Numerator of SSIM
    num_ssim = (2 * img_mat_mu_12 + c_1) * (2 * img_mat_sigma_12 + c_2)
    # Denominator of SSIM
    den_ssim = (img_mat_mu_1_sq + img_mat_mu_2_sq + c_1) *\
        (img_mat_sigma_1_sq + img_mat_sigma_2_sq + c_2)
    # SSIM
    ssim_map = num_ssim / den_ssim
    index = numpy.average(ssim_map)

    return index


# def msssim(img1, img2):
#     """This function implements Multi-Scale Structural Similarity (MSSSIM) Image
#     Quality Assessment according to Z. Wang's "Multi-scale structural similarity
#     for image quality assessment" Invited Paper, IEEE Asilomar Conference on
#     Signals, Systems and Computers, Nov. 2003

#     Author's MATLAB implementation:-
#     http://www.cns.nyu.edu/~lcv/ssim/msssim.zip
#     """
#     from scipy import ndimage
#     import numpy
#     level = 5
#     weight = numpy.array([0.0448, 0.2856, 0.3001, 0.2363, 0.1333])
#     downsample_filter = numpy.ones((2, 2)) / 4.0
#     im1 = img1.astype(numpy.float64)
#     im2 = img2.astype(numpy.float64)
#     mssim = numpy.array([])
#     mcs = numpy.array([])
#     for l in range(level):
#         ssim_map, cs_map = ssim(im1, im2, cs_map=True)
#         mssim = numpy.append(mssim, ssim_map.mean())
#         mcs = numpy.append(mcs, cs_map.mean())
#         filtered_im1 = ndimage.filters.convolve(im1, downsample_filter,
#                                                 mode='reflect')
#         filtered_im2 = ndimage.filters.convolve(im2, downsample_filter,
#                                                 mode='reflect')
#         im1 = filtered_im1[::2, ::2]
#         im2 = filtered_im2[::2, ::2]
#     return (numpy.prod(mcs[0:level - 1]**weight[0:level - 1]) *
#             (mssim[level - 1]**weight[level - 1]))


# def ssim(img1, img2, cs_map=False):
#     """Return the Structural Similarity Map corresponding to input images img1
#     and img2 (images are assumed to be uint8)

#     This function attempts to mimic precisely the functionality of ssim.m a
#     MATLAB provided by the author's of SSIM
#     https://ece.uwaterloo.ca/~z70wang/research/ssim/ssim_index.m
#     """
#     img1 = img1.astype(numpy.float64)
#     img2 = img2.astype(numpy.float64)
#     size = 11
#     sigma = 1.5
#     window = gauss.fspecial_gauss(size, sigma)
#     K1 = 0.01
#     K2 = 0.03
#     L = 255 #bitdepth of image
#     C1 = (K1*L)**2
#     C2 = (K2*L)**2
#     mu1 = signal.fftconvolve(window, img1, mode='valid')
#     mu2 = signal.fftconvolve(window, img2, mode='valid')
#     mu1_sq = mu1*mu1
#     mu2_sq = mu2*mu2
#     mu1_mu2 = mu1*mu2
#     sigma1_sq = signal.fftconvolve(window, img1*img1, mode='valid') - mu1_sq
#     sigma2_sq = signal.fftconvolve(window, img2*img2, mode='valid') - mu2_sq
#     sigma12 = signal.fftconvolve(window, img1*img2, mode='valid') - mu1_mu2
#     if cs_map:
#         return (((2*mu1_mu2 + C1)*(2*sigma12 + C2))/((mu1_sq + mu2_sq + C1)*
#                     (sigma1_sq + sigma2_sq + C2)),
#                 (2.0*sigma12 + C2)/(sigma1_sq + sigma2_sq + C2))
#     else:
#         return ((2*mu1_mu2 + C1)*(2*sigma12 + C2))/((mu1_sq + mu2_sq + C1)*
#                     (sigma1_sq + sigma2_sq + C2))



from tables import *
class AnalyticalPhantomMultiScaleReconstruction(IsDescription):
    phantom = Float32Col(shape=(256, 256))
    direct1x = Float32Col(shape=(256, 256))
    direct2x = Float32Col(shape=(256, 256))
    direct4x = Float32Col(shape=(256, 256))
    direct8x = Float32Col(shape=(256, 256))
    infer2x = Float32Col(shape=(256, 256))
    infer4x = Float32Col(shape=(256, 256))
    infer8x = Float32Col(shape=(256, 256))
    iterp2x = Float32Col(shape=(256, 256))
    iterp4x = Float32Col(shape=(256, 256))
    iterp8x = Float32Col(shape=(256, 256))
    direct1x = Float32Col(shape=(256, 256))
    direct2x = Float32Col(shape=(256, 256))
    direct4x = Float32Col(shape=(256, 256))
    direct8x = Float32Col(shape=(256, 256))
    infer2x = Float32Col(shape=(256, 256))
    infer4x = Float32Col(shape=(256, 256))
    infer8x = Float32Col(shape=(256, 256))
    iterp2x = Float32Col(shape=(256, 256))
    iterp4x = Float32Col(shape=(256, 256))
    iterp8x = Float32Col(shape=(256, 256))


import click

@click.command()
def hello():
    click.echo('HELLO FROM SR.HELLO!')

@click.command()
@click.option('--input', '-i', type=str)
@click.option('--output', '-o', type=str)
def analysis(input, output):
    from .analysis import analysis_result
    analysis_result(input, output)

@click.command()
@click.option('--input', '-i', type=str)
@click.option('--output', '-o', type=str)
def npz2mat(input, output):
    import numpy as np
    from scipy.io import savemat
    data = np.load(input)
    data = {k: data[k] for k in data}
    savemat(output, data)
    
# @click.command()
# @click.option('--input', '-i', type=str)
# @click.option('--output', '-o', type=str)
# @click.option('--metrices', '-m', type=str, )

@click.command()
@click.argument('sids', type=int, nargs=-1)
def squeue(sids):
    import os
    for sid in sids:
        cmd = 'squeue' 
        with os.popen(cmd) as fin:
            result = fin.readlines()



import click


@click.command()
@click.option('--input', '-i', type=str)
@click.option('--output', '-o', type=str)
@click.option('--keys', '-k', type=(str, str), multiple=True)
@click.option('--phantom_shape', '-s', type=int, nargs=2)
@click.option('--sen_width', '-w', type=float)
@click.option('--method', '-m', type=str, default='FBP')
@click.option('--nb_iter', '-n', type=int, default=1)
@click.option('--sino2pi', is_flag=True)
def reconnpz(input, output, keys, phantom_shape, sen_width, method, nb_iter, sino2pi):
    import numpy as np
    from dxpy.tensor.io import load_npz
    from tqdm import tqdm
    from dxpy.tensor.collections import dict_element_to_tensor
    from dxpy.medical_image_processing.phantom import Phantom2DSpec
    from dxpy.medical_image_processing.projection.parallel import projection2d
    from dxpy.medical_image_processing.reconstruction.parallel import reconstruction2d
    from dxpy.medical_image_processing.detector.parallel import Detector2DParallelRing
    sinos = load_npz(input)
    result = dict()
    nb_imgs = None
    nb_sen = 0
    if len(keys) == 0:
        keys_input = list()
        keys_output = list()
        for k in sinos:
            if k.startswith('sino_'):
                keys_input.append(k)
                keys_output.append(k.replace('sino_', 'recon_'))
    else:
        keys_input, keys_output = tuple(zip(*keys))
    keys_mapped = dict()
    for ki, ko in zip(keys_input, keys_output):
        keys_mapped[ki] = ko
    for k in sinos:
        if k not in keys_input:
            continue
        if sinos[k].shape[2] > nb_sen:
            nb_sen = sinos[k].shape[2]
    for k in sinos:
        if k not in keys_input:
            result[k] = sinos[k]
            continue
        if nb_imgs is None:
            nb_imgs = sinos[k].shape[0]
        else:
            if not nb_imgs == sinos[k].shape[0]:
                raise ValueError("Invalid sinos shape.")
        phan_spec = Phantom2DSpec(shape=phantom_shape)
        sen_width_c = sen_width * nb_sen / sinos[k].shape[2]
        nb_views = sinos[k].shape[1]
        if sino2pi:
            nb_views //= 2
        views = np.linspace(0, np.pi, nb_views, endpoint=False)
        detector = Detector2DParallelRing(nb_sensors=sinos[k].shape[2],
                                          sensor_width=sen_width_c,
                                          views=views)
        print('I: Reconstructing: {}.'.format(k))
        ko = keys_mapped[k]
        result[ko] = []
        for i in tqdm(range(nb_imgs)):
            sino = sinos[k][i, ...]
            if sino2pi:
                sino = sino[:sino.shape[0]//2, :] 
            recon = reconstruction2d(sino, detector, phan_spec,
                                     method=method,
                                     iterations=nb_iter)
            result[ko].append(recon)
    dict_element_to_tensor(result)
    np.savez(output, **result)



import numpy as np
from typing import TypeVar


def _create_2d_ndarray(images, shape=None):
    if shape is None:
        shape = images.shape[:2]
    result = np.empty(shape[:2], np.ndarray)
    if isinstance(images, np.ndarray):
        images = images.tolist()
    for i in range(shape[0]):
        for j in range(shape[1]):
            if images[i] is None or images[i][j] is None:
                result[i, j] = None
            else:
                result[i, j] = np.array(images[i][j])
    return result


def _handle_list_inputs(images, error):
    if isinstance(images, (list, tuple)):
        types = {type(im) for im in images}
        types.discard(None)
        if len(types) > 1:
            raise TypeError('images have different types: {}.'.format(types))
        types = list(types)
        if types[0] == np.ndarray:
            nc = min([im.shape[0] for im in images if im is not None])
        elif types[0] in (list, tuple):
            nc = min([len(im) for im in images if im is not None])
        else:
            raise TypeError("Type {} not supported.".format(types[0]))
        nr = len(images)
        return _create_2d_ndarray(images, [nr, nc])
    else:
        raise error


def _unified_images(images, invert_row_column):
    try:
        images = np.array(images)
    except ValueError as e:
        images = _handle_list_inputs(images, e)
    if images.ndim == 1:
        images = _handle_list_inputs(images.tolist(), ValueError(
            "Failed to analysis, hint: images.dim==1."))
    if images.ndim > 2:
        images = _create_2d_ndarray(images)
    if invert_row_column:
        images = np.moveaxis(images, 1, 0)
    # Unified array to 2 dimension
    for ir in range(images.shape[0]):
        for ic in range(images.shape[1]):
            t = images[ir, ic]
            if t is None:
                continue
            if t.ndim == 2:
                continue
            if t.ndim == 3 and t.shape[-1] == 3:
                continue
            if t.ndim == 4:
                # Case [1, M, N, 1]
                if t.shape[0] == 1 and t.shape[3] == 1:
                    images[ir, ic] = t[0, :, :, 0]
                    continue
                # Case [1, M, N, 3]
                if t.shape[0] == 1 and t.shape[3] == 3:
                    images[ir, ic] = t[0, :, :, :]
                    continue
            elif t.ndim == 3:
                # Case [M, N, 1]
                if t.shape[-1] == 1:
                    images[ir, ic] = t[:, :, 0]
                    continue
                # Case [1, M, N]
                if t.shape[0] == 1:
                    images[ir, ic] = t[0, :, :]
                    continue
            raise ValueError(
                'Invalid images[{}, {}] with shape {}.'.format(ir, ic, t.shape))
    return images


def _auto_windows(images, windows):
    windows = np.zeros(list(images.shape) + [2])
    for i in range(images.shape[0]):
        for j in range(images.shape[1]):
            if images[i, j] is None:
                windows[i, j] = np.nan
            windows[i, j, 0] = np.min(images[i, j])
            windows[i, j, 1] = np.max(images[i, j])
    return windows


def _unified_windows(images, windows):
    if windows is None:
        return _auto_windows(images, windows)
    windows = np.array(windows)
    if windows.ndim < 3:
        windows = np.reshape(
            windows, [1] * (3 - windows.ndim) + list(windows.shape))
    # Force broadcast windows shape to [R,C,2]
    dummy_zero = np.zeros(list(images.shape[:2]) + [2])
    windows = windows + dummy_zero
    return windows


def _adjust_images_to_fit_nb_columns(images, windows, max_columns=None, max_rows=None):
    """
    images: np.array of shape [R, C]
    windows: np.array of shape [R, C, 2]
    """
    if not images.shape == windows.shape[:2]:
        raise ValueError("Images shape {} and windows shape {} not consist.".format(
            images.shape, windows.shape))
    nb_row_pre, nb_column_pre = images.shape
    if max_columns is not None and nb_column_pre > max_columns:
        nb_ratio = np.ceil(double(nb_column_pre) / max_columns)
        nb_row = nb_row_pre * nb_ratio
        new_images = np.empty([nb_row, max_columns], np.ndarray)
        new_windows = np.empty([nb_row, max_columns, 2], np.ndarray)
        for i in range(nb_ratio):
            r_s, r_e = (i * nb_row_pre, (i + 1) * nb_row_pre)
            c_s, c_e = (i * max_columns,
                        min((i + 1) * max_columns, nb_column_pre))
            new_images[r_s: r_e, :] = images[:, c_s: c_e]
            new_windows[r_s: r_e, :, :] = windows[:, c_s: c_e, :]
        images = new_images
        windows = new_windows
    if max_rows is not None and images.shape[0] > max_rows:
        images = images[:max_rows, :]
        windows = windows[:max_rows, :]
    return images, windows


def _adjust_figure_size(images, scale, scale_factor):
    # Default scale: 10: 100
    width = 0.0
    height = 0.0
    nb_images = 0
    for ir in range(images.shape[0]):
        for ic in range(images.shape[1]):
            if images[ir, ic] is None:
                continue
            width += images[ir, ic].shape[1]
            height += images[ir, ic].shape[0]
            nb_images += 1
    width = width / nb_images * images.shape[1]
    height = height / nb_images * images.shape[0]
    DEFAULT_WIDTH = scale_factor * scale
    ratio = DEFAULT_WIDTH / width
    height = height * ratio
    return (DEFAULT_WIDTH, height), 100.0 
    # return (images.shape[1] + 1, images.shape[0] + 1), 640.0 / (images.shape[1] + 1)


def grid_view(images, windows=None, scale=1.0, cmap=None,
              *,
              max_columns=None, max_rows=None,
              hide_axis=True,
              hspace=0.1, wspace=0.1, return_figure=False, dpi=None, adjust_figure_size=True, save_filename=None,
              invert_row_column=False,
              scale_factor=2.0,
              _top=None,
              _right=None):
    import matplotlib.pyplot as plt
    import matplotlib.gridspec as gridspec
    from matplotlib.figure import SubplotParams
    """ subplot list of images of multiple categories into grid subplots

    Args:
        image_lists: list of [list of images or 4D tensor]
        windows: list of windows
        nb_column: columns of images
    Returns:
        Return figure if return_figure is true, else None.
    """
    images = _unified_images(images, invert_row_column)
    windows = _unified_windows(images, windows)
    images, windows = _adjust_images_to_fit_nb_columns(images, windows,
                                                       max_columns, max_rows)
    figsize, default_dpi = _adjust_figure_size(images, scale, scale_factor)
    from dxpy.debug.utils import dbgmsg
    dbgmsg(figsize, default_dpi)
    dbgmsg(images.shape)
    if dpi is None:
        dpi = default_dpi
    dpi = dpi * scale
    fig = plt.figure(figsize=figsize, dpi=dpi,
                     subplotpars=SubplotParams(left=0.0, right=1.0, bottom=0.0, top=1.0, wspace=0.0, hspace=0.0))
    # fig.subplots_adjust(hspace=hspace, wspace=wspace)
    nr, nc = images.shape
    if _top is None:
        # _top = figsize[1] / nr * nc
        _top = scale
    if _right is None:
        _right = figsize[0] / nc * nr
        _right = scale
    gs = gridspec.GridSpec(nr, nc,
                           wspace=wspace, hspace=hspace,
                           top=_top, bottom=0.0,
                           left=0.0, right=_right)
    for ir in range(nr):
        for ic in range(nc):
            if images[ir, ic] is None:
                continue
            # ax = plt.subplot(nr, nc, ir * nc + ic + 1)
            ax = plt.subplot(gs[ir, ic])
            ax.imshow(images[ir, ic], cmap=cmap,
                      vmin=windows[ir, ic, 0], vmax=windows[ir, ic, 1])
            if hide_axis:
                plt.axis('off')
            else:
                ax.set_xticklabels([])
                ax.set_yticklabels([])
    if save_filename is not None:
        fig.savefig(save_filename)
    if return_figure:
        return fig



def dict_append(target, new_item):
    if not isinstance(target, dict):
        raise TypeError(
            'Target of dict_append needs to be dict, got {}.'.format(type(target)))
    for k in target:
        target[k].append(new_item[k])


def dict_element_to_tensor(target):
    import numpy as np
    if not isinstance(target, dict):
        raise TypeError(
            'Target of dict_append needs to be dict, got {}.'.format(type(target)))
    for k in target:
        target[k] = np.array(target[k])



import click


@click.group()
def npz():
    pass


@npz.command()
@click.argument('filename')
def ls(filename):
    from ..npz import list_npz
    list_npz(filename)

@npz.command()
@click.argument('filename')
@click.option('--output', '-o', type=str, help='Outputname, use same name and change suffix to mat if not given.')
def tomat(filename, output):
    from ..npz import npz2mat
    if output is None:
        output = filename[:-3] + 'mat'
    npz2mat(filename, output)

@npz.command()
@click.argument('filename', nargs=-1)
@click.argument('output')    
def cat(filename, output):
    from ..npz import concatenate
    concatenate(filename, output)




import click
from tqdm import tqdm


def infer_sino_sr(dataset, nb_samples, output):
    """
    Use network in current directory as input for inference
    """
    import tensorflow as tf
    from dxpy.learn.dataset.api import get_dataset
    from dxpy.learn.net.api import get_network
    from dxpy.configs import ConfigsView
    from dxpy.learn.config import config
    import numpy as np
    import yaml
    from dxpy.debug.utils import dbgmsg
    dbgmsg(dataset)
    data_raw = np.load(dataset)
    data_raw = {k: np.array(data_raw[k]) for k in data_raw.keys()}
    config_view = ConfigsView(config)

    def tensor_shape(key):
        shape_origin = data_raw[key].shape
        return [1] + list(shape_origin[1:3]) + [1]
    with tf.name_scope('inputs'):
        keys = ['input/image{}x'.format(2**i) for i in range(4)]
        keys += ['label/image{}x'.format(2**i) for i in range(4)]
        dataset = {k: tf.placeholder(
            tf.float32, tensor_shape(k)) for k in keys}

    network = get_network('network/srms', dataset=dataset)
    nb_down_sample = network.param('nb_down_sample')
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    sess = tf.train.MonitoredTrainingSession(
        checkpoint_dir='./save', config=config, save_checkpoint_secs=None)

    STAT_STD = 9.27
    STAT_MEAN = 9.76
    BASE_SHAPE = (640, 320)

    dataset_configs = config_view['dataset']['srms']
    with_noise = dataset_configs['with_poission_noise']
    if with_noise:
        PREFIX = 'input'
    else:
        PREFIX = 'label'

    def crop_sinogram(tensor, target_shape=None):
        if target_shape is None:
            target_shape = BASE_SHAPE
        if len(tensor.shape) == 4:
            tensor = tensor[0, :, :, 0]
        o1 = (tensor.shape[0] - target_shape[0]) // 2
        o2 = (tensor.shape[1] - target_shape[1]) // 2
        return tensor[o1:-o1, o2:-o2]

    def run_infer(idx):
        input_key = '{}/image{}x'.format(PREFIX, 2**nb_down_sample)
        low_sino = np.reshape(
            data_raw[input_key][idx, :, :], tensor_shape(input_key))
        low_sino = (low_sino - STAT_MEAN) / STAT_STD
        feeds = {dataset['input/image{}x'.format(2**nb_down_sample)]: low_sino}
        inf, itp = sess.run([network['outputs/inference'],
                             network['outputs/interp']], feed_dict=feeds)
        infc = crop_sinogram(inf)
        itpc = crop_sinogram(itp)
        infc = infc * STAT_STD + STAT_MEAN
        itpc = itpc * STAT_STD + STAT_MEAN
        return infc, itpc

    phans = []
    sino_highs = []
    sino_lows = []
    sino_itps = []
    sino_infs = []
    NB_MAX = data_raw['phantom'].shape[0]
    for idx in tqdm(range(nb_samples), ascii=True):
        if idx > NB_MAX:
            import sys
            print('Index {} out of range {}, stop running and store current result...'.format(
                idx, NB_MAX), file=sys.stderr)
            break

        phans.append(data_raw['phantom'][idx, ...])
        sino_highs.append(crop_sinogram(
            data_raw['{}/image1x'.format(PREFIX)][idx, :, :]))
        sino_lows.append(crop_sinogram(data_raw['{}/image{}x'.format(
            PREFIX, 2**nb_down_sample)][idx, ...], [s // (2**nb_down_sample) for s in BASE_SHAPE]))
        sino_inf, sino_itp = run_infer(idx)
        sino_infs.append(sino_inf)
        sino_itps.append(sino_itp)

    results = {'phantom': phans, 'sino_itps': sino_itps,
               'sino_infs': sino_infs, 'sino_highs': sino_highs, 'sino_lows': sino_lows}
    np.savez(output, **results)


def infer_phan_sr(dataset, nb_samples, output):
    """
    Use network in current directory as input for inference
    """
    import tensorflow as tf
    from dxpy.learn.dataset.api import get_dataset
    from dxpy.learn.net.api import get_network
    from dxpy.configs import ConfigsView
    from dxpy.learn.config import config
    import numpy as np
    import yaml
    from dxpy.debug.utils import dbgmsg
    print('Using dataset file:', dataset)
    data_raw = np.load(dataset)
    data_raw = {k: np.array(data_raw[k]) for k in data_raw.keys()}
    config_view = ConfigsView(config)

    dataset_configs = config_view['dataset']['srms']
    with_noise = dataset_configs['with_poission_noise']
    dbgmsg(with_noise)
    if with_noise:
        PREFIX = 'noise'
    else:
        PREFIX = 'clean'

    def data_key(nd):
        return '{}{}x'.format(PREFIX, 2**nd)

    def tensor_shape(key):
        shape_origin = data_raw[key].shape
        return [1] + list(shape_origin[1:3]) + [1]
    with tf.name_scope('inputs'):
        keys = ['input/image{}x'.format(2**i) for i in range(4)]
        keys += ['label/image{}x'.format(2**i) for i in range(4)]
        dataset = {k: tf.placeholder(tf.float32, tensor_shape(data_key(i % 4)))
                   for i, k in enumerate(keys)}

    network = get_network('network/srms', dataset=dataset)
    nb_down_sample = network.param('nb_down_sample')
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    sess = tf.train.MonitoredTrainingSession(
        checkpoint_dir='./save', config=config, save_checkpoint_secs=None)

    STAT_MEAN = 15.26
    STAT_STD = 22.0
    STAT_MEAN_LOW = 15.26 * (4.0**nb_down_sample)
    STAT_STD_LOW = 22.0 * (4.0**nb_down_sample)
    BASE_SHAPE = (256, 256)

    def crop_image(tensor, target_shape=None):
        if target_shape is None:
            target_shape = BASE_SHAPE
        if len(tensor.shape) == 4:
            tensor = tensor[0, :, :, 0]
        o1 = (tensor.shape[0] - target_shape[0]) // 2
        o2 = (tensor.shape[1] - target_shape[1]) // 2
        return tensor[o1:-o1, o2:-o2]

    input_key = data_key(nb_down_sample)
    dbgmsg('input_key:', input_key)

    def run_infer(idx):
        low_phan = np.reshape(
            data_raw[input_key][idx, :, :], tensor_shape(input_key))
        # low_phan = (low_phan - STAT_MEAN) / STAT_STD
        feeds = {dataset['input/image{}x'.format(2**nb_down_sample)]: low_phan}
        inf, itp = sess.run([network['outputs/inference'],
                             network['outputs/interp']], feed_dict=feeds)
        infc = crop_image(inf)
        itpc = crop_image(itp)
        infc = infc * STAT_STD_LOW + STAT_MEAN_LOW
        itpc = itpc * STAT_STD_LOW + STAT_MEAN_LOW
        return infc, itpc

    phans = []
    img_highs = []
    img_lows = []
    img_itps = []
    img_infs = []
    NB_MAX = data_raw['phantom'].shape[0]
    for idx in tqdm(range(nb_samples), ascii=True):
        if idx > NB_MAX:
            import sys
            print('Index {} out of range {}, stop running and store current result...'.format(
                idx, NB_MAX), file=sys.stderr)
            break

        phans.append(data_raw['phantom'][idx, ...])
        img_high = crop_image(data_raw[data_key(0)][idx, :, :])
        img_high = img_high * STAT_STD + STAT_MEAN
        # img_high = img_high * STAT_STD / \
        # (4.0**nb_down_sample) + STAT_MEAN / (4.0**nb_down_sample)
        img_highs.append(img_high)
        img_low = crop_image(data_raw[data_key(nb_down_sample)][idx, ...],
                             [s // (2**nb_down_sample) for s in BASE_SHAPE])
        img_low = img_low * STAT_STD_LOW + STAT_MEAN_LOW
        img_lows.append(img_low)
        img_inf, img_itp = run_infer(idx)
        img_infs.append(img_inf)
        img_itps.append(img_itp)

    img_highs = np.array(img_highs)
    img_infs = np.array(img_infs) / (4.0**nb_down_sample)
    img_itps = np.array(img_itps) / (4.0**nb_down_sample)
    img_lows = np.array(img_lows) / (4.0**nb_down_sample)

    results = {'phantom': phans, 'img_itps': img_itps,
               'img_infs': img_infs, 'img_highs': img_highs, 'img_lows': img_lows}
    np.savez(output, **results)


def infer_mct(dataset, nb_samples, output):
    """
    Use network in current directory as input for inference
    """
    import tensorflow as tf
    from dxpy.learn.dataset.api import get_dataset
    from dxpy.learn.net.api import get_network
    from dxpy.configs import ConfigsView
    from dxpy.learn.config import config
    import numpy as np
    import yaml
    from dxpy.debug.utils import dbgmsg
    print('Using dataset file:', dataset)
    data_raw = np.load(dataset)
    data_raw = {k: np.array(data_raw[k]) for k in data_raw.keys()}
    config_view = ConfigsView(config)

    def data_key(nd):
        return 'image{}x'.format(2**nd)

    def tensor_shape(key):
        shape_origin = data_raw[key].shape
        return [1] + list(shape_origin[1:3]) + [1]

    with tf.name_scope('inputs'):
        keys = ['input/image{}x'.format(2**i) for i in range(4)]
        keys += ['label/image{}x'.format(2**i) for i in range(4)]
        dataset = {k: tf.placeholder(tf.float32, tensor_shape(data_key(i % 4)))
                   for i, k in enumerate(keys)}

    network = get_network('network/srms', dataset=dataset)
    nb_down_sample = network.param('nb_down_sample')
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    sess = tf.train.MonitoredTrainingSession(
        checkpoint_dir='./save', config=config, save_checkpoint_secs=None)

    STAT_MEAN = 9.93
    STAT_STD = 7.95
    STAT_MEAN_LOW = 9.93 * (4.0**nb_down_sample)
    STAT_STD_LOW = 7.95 * (4.0**nb_down_sample)
    BASE_SHAPE = (384, 384)

    def crop_image(tensor, target_shape=None):
        if target_shape is None:
            target_shape = BASE_SHAPE
        if len(tensor.shape) == 4:
            tensor = tensor[0, :, :, 0]
        # o1 = (tensor.shape[0] - target_shape[0]) // 2
        o1 = tensor.shape[0] // 2
        o2 = (tensor.shape[1] - target_shape[1]) // 2
        return tensor[o1:o1 + target_shape[0], o2:-o2]

    input_key = data_key(nb_down_sample)
    dbgmsg('input_key:', input_key)

    def run_infer(idx):
        low_phan = np.reshape(
            data_raw[input_key][idx, :, :], tensor_shape(input_key))
        # low_phan = (low_phan - STAT_MEAN) / STAT_STD
        feeds = {dataset['input/image{}x'.format(2**nb_down_sample)]: low_phan}
        inf, itp = sess.run([network['outputs/inference'],
                             network['outputs/interp']], feed_dict=feeds)
        infc = crop_image(inf)
        itpc = crop_image(itp)
        infc = infc * STAT_STD_LOW + STAT_MEAN_LOW
        itpc = itpc * STAT_STD_LOW + STAT_MEAN_LOW
        return infc, itpc

    phans = []
    img_highs = []
    img_lows = []
    img_itps = []
    img_infs = []
    NB_MAX = data_raw['phantom'].shape[0]
    for idx in tqdm(range(nb_samples), ascii=True):
        if idx > NB_MAX:
            import sys
            print('Index {} out of range {}, stop running and store current result...'.format(
                idx, NB_MAX), file=sys.stderr)
            break

        phans.append(data_raw['phantom'][idx, ...])
        img_high = crop_image(data_raw[data_key(0)][idx, :, :])
        img_high = img_high * STAT_STD + STAT_MEAN
        # img_high = img_high * STAT_STD / \
        # (4.0**nb_down_sample) + STAT_MEAN / (4.0**nb_down_sample)
        img_highs.append(img_high)
        img_low = crop_image(data_raw[data_key(nb_down_sample)][idx, ...],
                             [s // (2**nb_down_sample) for s in BASE_SHAPE])
        img_low = img_low * STAT_STD_LOW + STAT_MEAN_LOW
        img_lows.append(img_low)
        img_inf, img_itp = run_infer(idx)
        img_infs.append(img_inf)
        img_itps.append(img_itp)

    img_highs = np.array(img_highs)
    img_infs = np.array(img_infs) / (4.0**nb_down_sample)
    img_itps = np.array(img_itps) / (4.0**nb_down_sample)
    img_lows = np.array(img_lows) / (4.0**nb_down_sample)

    results = {'phantom': phans, 'sino_itps': img_itps,
               'sino_infs': img_infs, 'sino_highs': img_highs, 'sino_lows': img_lows}
    np.savez(output, **results)


def recon_sino(sinograms_filename, nb_samples, output, recon_method):
    import numpy as np
    from dxpy.medical_image_processing.phantom import Phantom2DSpec
    from dxpy.medical_image_processing.reconstruction.parallel import reconstruction2d
    from dxpy.medical_image_processing.detector import Detector2DParallelRing
    sinograms = np.load(sinograms_filename)
    sino_highs = np.array(sinograms['sino_highs'])
    sino_lows = np.array(sinograms['sino_lows'])
    sino_infs = np.array(sinograms['sino_infs'])
    sino_itps = np.array(sinograms['sino_itps'])
    phans = np.array(sinograms['phantom'])
    nb_views_high = sino_highs.shape[1]
    nb_views_low = sino_lows.shape[1]
    nb_views_high //= 2
    nb_views_low //= 2
    sino_highs = sino_highs[:, :nb_views_high, :]
    sino_lows = sino_lows[:, :nb_views_low, :]
    sino_infs = sino_infs[:, :nb_views_high, :]
    sino_itps = sino_itps[:, :nb_views_high, :]
    sino_lows = sino_lows * nb_views_high**2 / (nb_views_low**2)
    phan_spec = Phantom2DSpec(shape=[256] * 2)
    detec_high = Detector2DParallelRing(views=np.linspace(
        0, np.pi, nb_views_high, endpoint=False), nb_sensors=nb_views_high, sensor_width=1.0)
    detec_low = Detector2DParallelRing(views=np.linspace(
        0, np.pi, nb_views_low, endpoint=False), nb_sensors=nb_views_low, sensor_width=1.0 * nb_views_high / nb_views_low)
    recon_highs = []
    recon_lows = []
    recon_infs = []
    recon_itps = []

    def recon_kernel(sinogram, detec):
        sinogram = np.maximum(sinogram, 0.0)
        if recon_method == 'fbp':
            recon = reconstruction2d(sinogram, detec, phan_spec)
        elif recon_method == 'sart':
            from dxpy.debug.utils import dbgmsg
            dbgmsg('USING SART !!!!!!!!!!')
            recon = reconstruction2d(
                sinogram, detec, phan_spec, method='SART_CUDA', iterations=500)
        recon = np.maximum(recon, 0.0)
        recon = recon / np.sum(recon) * 1e6
        return recon

    for i in tqdm(range(nb_samples), ascii=True):
        recon_highs.append(recon_kernel(sino_highs[i, ...], detec_high))
        recon_lows.append(recon_kernel(sino_lows[i, ...], detec_low))
        recon_infs.append(recon_kernel(sino_infs[i, ...], detec_high))
        recon_itps.append(recon_kernel(sino_itps[i, ...], detec_high))
    results = {'phantom': phans, 'recon_highs': recon_highs,
               'recon_lows': recon_lows, 'recon_infs': recon_infs, 'recon_itps': recon_itps}
    np.savez(output, **results)


def _update_statistics(dataset_config_name, low_dose_ratio):
    from dxpy.learn.config import config
    from dxpy.core.path import Path
    if isinstance(dataset_config_name, str):
        dataset_config_name = Path(dataset_config_name).parts()
    c = config
    for k in dataset_config_name:
        c = c[k]
    mean, std = c['mean'], c['std']
    mean /= low_dose_ratio
    std /= low_dose_ratio
    c['mean'], c['std'] = mean, std
    return mean, std


def _load_input_and_phantom(input_filename, phantom_filename, input_key, phantom_key):
    from dxpy.tensor.io import load_npz
    phantom_filename = phantom_filename or input_filename
    data_input = load_npz(input_filename)
    if phantom_filename == input_filename:
        data_phantom = data_input
    else:
        data_phantom = load_npz(phantom_filename)
    return data_input[input_key], data_phantom[phantom_key]


def _get_input_dict(dataset):
    nb_down = dataset.param('nb_down_sample')
    nb_down_ratio = [2**i for i in range(nb_down + 1)]
    prefix = 'noise' if dataset.param('with_poission_noise') else 'clean'
    label_key = '{}/image1x'.format(prefix)
    input_key = '{}/image{}x'.format(prefix, 2**nb_down)
    input_dict = dict()
    for nd in nb_down_ratio:
        data_current_scale = dataset['{}/image{}x'.format(prefix, nd)]
        input_dict['input/image{}x'.format(nd)] = data_current_scale
        input_dict['label/image{}x'.format(nd)] = data_current_scale
    return input_dict


def infer_ext(input_npz_filename, input_key='clean/image1x',
              phantom_npz_filename=None, phantom_key='phantom',
              dataset_config_name='dataset/srms',
              network_config_name='network/srms',
              output_shape=(320, 320),
              output='infer_ext_result.npz',
              ids=None, nb_run=1, low_dose_ratio=1.0,
              crop_method='hc',
              config_filename='dxln.yml'):
    """
    """
    import numpy as np
    from dxpy.learn.dataset.api import get_dataset
    from dxpy.learn.net.api import get_network
    from dxpy.tensor.collections import dict_append, dict_element_to_tensor
    from dxpy.learn.utils.general import load_yaml_config, pre_work
    from tqdm import tqdm
    from dxpy.learn.session import SessionMonitored
    pre_work()
    inputs, phantoms = _load_input_and_phantom(input_npz_filename,
                                               phantom_npz_filename,
                                               input_key, phantom_key)
    data_input = inputs / low_dose_ratio
    load_yaml_config(config_filename)
    mean, std = _update_statistics(dataset_config_name, low_dose_ratio)
    dataset = get_dataset(dataset_config_name)
    dataset_feed = dataset['external_place_holder']
    nb_down = dataset.param('nb_down_sample')
    nb_down_ratio = [2**i for i in range(nb_down + 1)]
    prefix = 'noise' if dataset.param('with_poission_noise') else 'clean'
    label_key = '{}/image1x'.format(prefix)
    input_key = '{}/image{}x'.format(prefix, 2**nb_down)
    network = get_network(network_config_name,
                          dataset=_get_input_dict(dataset))
    sess = SessionMonitored()
    fetches = {'label': dataset[label_key],
               'input': dataset[input_key],
               'infer': network['inference'],
               'interp': network['outputs/interp']}

    def proc(result, is_low=False):
        from dxpy.tensor.transform import crop_to_shape
        if is_low:
            target_shape = [s // (2**nb_down) for s in output_shape]
        else:
            target_shape = output_shape
        result = crop_to_shape(result, target_shape, '0' + crop_method + '0')
        result = result * std + mean
        result = np.maximum(result, 0.0)
        return result

    def get_result(idx):
        phan = phantoms[idx, ...]
        result = sess.run(fetches, feed_dict={
                          dataset_feed: np.reshape(data_input[idx:idx + 1, ...],
                                                   dataset.param('input_shape'))})
        result_c = {'sino_highs': proc(result['label']),
                    'sino_lows': proc(result['input'], True),
                    'sino_infs': proc(result['infer']),
                    'sino_itps': proc(result['interp']),
                    'phantoms': phan}
        return result_c
    keys = ['sino_highs', 'sino_lows', 'sino_infs', 'sino_itps', 'phantoms']
    results_sino = {k: [] for k in keys}
    for idx in tqdm(ids, ascii=True):
        for _ in tqdm(range(nb_run), ascii=True, leave=False):
            result_sino = get_result(idx)
            dict_append(results_sino, result_sino)
    dict_element_to_tensor(results_sino)
    np.savez(output, **results_sino)


def infer_mice(dataset, nb_samples, output):
    import numpy as np
    import tensorflow as tf
    from dxpy.learn.dataset.api import get_dataset
    from dxpy.learn.net.api import get_network
    from dxpy.learn.utils.general import load_yaml_config, pre_work
    from dxpy.learn.session import Session
    from dxpy.numpy_extension.visual import grid_view
    from tqdm import tqdm
    pre_work()
    input_data = np.load(
        '/home/hongxwing/Workspace/NetInference/Mice/mice_test_data.npz')
    input_data = {k: np.array(input_data[k]) for k in input_data}
    dataset_origin = get_dataset('dataset/srms')
    is_low_dose = dataset_origin.param('low_dose')
    from dxpy.debug.utils import dbgmsg
    dbgmsg('IS LOW DOSE: ', is_low_dose)
    for k in input_data:
        print(k, input_data[k].shape)
    input_keys = ['input/image{}x'.format(2**i) for i in range(4)]
    label_keys = ['label/image{}x'.format(2**i) for i in range(4)]
    shapes = [[1] + list(input_data['clean/image{}x'.format(2**i)
                                    ].shape)[1:] + [1] for i in range(4)]
    inputs = {input_keys[i]: tf.placeholder(
        tf.float32, shapes[i], name='input{}x'.format(2**i)) for i in range(4)}
    labels = {label_keys[i]: tf.placeholder(
        tf.float32, shapes[i], name='label{}x'.format(2**i)) for i in range(4)}
    dataset = dict(inputs)
    dataset.update(labels)
    network = get_network('network/srms', dataset=dataset)
    nb_down_sample = network.param('nb_down_sample')
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    sess = tf.train.MonitoredTrainingSession(
        checkpoint_dir='./save', config=config, save_checkpoint_secs=None)

    if not is_low_dose:
        prefix = 'clean/image'
    else:
        prefix = 'noise/image'

    def get_feed(idx):
        #     return dict()
        data_raw = input_data['{}{}x'.format(
            prefix, 2**nb_down_sample)][idx, ...]
        data_raw = np.reshape(data_raw, [1] + list(data_raw.shape) + [1])
        data_label = input_data['{}1x'.format(prefix)][idx, ...]
        data_label = np.reshape(data_label, [1] + list(data_label.shape) + [1])
        return {dataset['input/image{}x'.format(2**nb_down_sample)]: data_raw,
                dataset['input/image1x'.format(2**nb_down_sample)]: data_label,
                dataset['label/image1x'.format(2**nb_down_sample)]: data_label, }

    to_run = {
        'inf': network['outputs/inference'],
        'itp': network['outputs/interp'],
        'high': network['input/image1x'],
        #     'li': network['outputs/loss_itp'],
        #     'ls': network['outputs/loss'],
        #     'la': network['outputs/aligned_label']
    }

    def crop(data, target):
        if len(data.shape) == 4:
            data = data[0, :, :, 0]
        o1 = data.shape[0] // 2
        o2 = (data.shape[1] - target[1]) // 2
        return data[o1:o1 + target[0], o2:-o2]

    MEAN = 100.0
    STD = 150.0
    if is_low_dose:
        MEAN /= dataset_origin.param('low_dose_ratio')
        STD /= dataset_origin.param('low_dose_ratio')
    NB_IMAGES = nb_samples

    def get_infer(idx):
        result = sess.run(to_run, feed_dict=get_feed(idx))
        inf = crop(result['inf'], [320, 64])
        itp = crop(result['itp'], [320, 64])
        high = crop(input_data['{}1x'.format(prefix)][idx, ...], [320, 64])
        low = crop(input_data['{}{}x'.format(prefix, 2**nb_down_sample)]
                   [idx, ...], [320 // (2**nb_down_sample), 64 // (2**nb_down_sample)])

        high = high * STD + MEAN
        low = low * STD + MEAN
        inf = inf * STD + MEAN
        itp = itp * STD + MEAN
        high = np.pad(high, [[0, 0], [32, 32]], mode='constant')
        low = np.pad(
            low, [[0, 0], [32 // (2**nb_down_sample)] * 2], mode='constant')
        inf = np.pad(inf, [[0, 0], [32, 32]], mode='constant')
        # inf = np.maximum(inf, 0.0)
        itp = np.pad(itp, [[0, 0], [32, 32]], mode='constant')
        return high, low, inf, itp

    results = {'high': [], 'low': [], 'inf': [], 'itp': []}
    for i in tqdm(range(NB_IMAGES)):
        high, low, inf, itp = get_infer(i)
        results['high'].append(high)
        results['low'].append(low)
        results['inf'].append(inf)
        results['itp'].append(itp)

    np.savez(output, **results)



def npz2mat(input_filename, output_filename):
    from scipy.io import savemat
    from .io import load_npz
    savemat(output, load_npz(input_filename))


def list_npz(input_filename):
    import numpy as np
    data = np.load(input_filename)
    for k in data:
        print(k, data[k].shape)


def concatenate(filenames, output_filename):
    from .io import load_npz
    import numpy as np
    result = None
    for f in filenames:
        data = load_npz(f)
        if result is None:
            result = data
        else:
            for k in result:
                result[k] = np.concatenate([result[k], data[k]], axis=0)
    np.savez(output_filename, **result)



import tensorflow as tf
from dxpy.configs import configurable
from dxpy.learn.graph import Graph, NodeKeys
from dxpy.learn.config import config
from dxpy.learn.utils.tensor import shape_as_list
import numpy as np
from typing import List


class SuperResolutionExternalDataset(Graph):
    @configurable(config, with_name=True)
    def __init__(self,
                 name='dataset/external',
                 nb_down_sample: int=3,
                 input_shape: List[int]=None,
                 with_random_crop: bool=True,
                 with_white_normalization: bool=True,
                 with_poission_noise: bool=False,
                 with_noise_label: bool=False,
                 target_shape: List[int]=None,
                 mean: float=0.0,
                 std: float=1.0,
                 **kw):
        super().__init__(name,
                         with_poission_noise=with_poission_noise,
                         target_shape=target_shape,
                         with_white_normalization=with_white_normalization,
                         with_noise_label=with_noise_label,
                         with_random_crop=with_random_crop,
                         nb_down_sample=nb_down_sample,
                         input_shape=input_shape,
                         mean=mean,
                         std=std,
                         **kw)
        result = self._get_sinogram_external()
        for k in result:
            self.register_node(k, result[k])

    def _nb_downs(self, start=0, offset=0, *, is_ratio=False):
        result = range(start, self.param('nb_down_sample') + offset)
        if is_ratio:
            result = [2**i for i in result]
        else:
            result = list(result)
        return result

    def _make_input_place_holder(self):
        with tf.name_scope('input'):
            dataset = tf.placeholder(tf.float32,
                                     self.param('input_shape'),
                                     name='external_input')
            return dataset

    def _target_shape(self, dataset):
        batch_size = shape_as_list(dataset[0])
        return [batch_size] + list(self.param('target_shape')) + [1]

    def _get_sinogram_external(self):
        from dxpy.learn.model.normalizer import FixWhite
        from ..super_resolution import SuperResolutionDataset
        with tf.name_scope('external_dataset'):
            dataset = self._make_input_place_holder()
            self.register_node('external_place_holder', dataset)
            if self.param('with_poission_noise'):
                with tf.name_scope('add_with_poission_noise'):
                    noise = tf.random_poisson(dataset, shape=[])
                    dataset = tf.concat([noise, dataset], axis=0)
            if self.param('with_white_normalization'):
                dataset = FixWhite(name=self.name / 'fix_white',
                                   inputs=dataset,
                                   mean=self.param('mean'),
                                   std=self.param('std')).as_tensor()
            if self.param('with_random_crop'):
                dataset = tf.random_crop(dataset, self._target_shape(dataset))
            dataset = SuperResolutionDataset(self.name / 'super_resolution',
                                             lambda: {'image': dataset},
                                             input_key='image',
                                             nb_down_sample=self.param('nb_down_sample'))
            keys = ['image{}x'.format(2**i)
                    for i in range(dataset.param('nb_down_sample') + 1)]
            result = dict()
            if self.param('with_poission_noise'):
                result.update(
                    {'noise/' + k: dataset[k][:shape_as_list(dataset[k])[0] // 2, ...] for k in keys})
                result.update(
                    {'clean/' + k: dataset[k][shape_as_list(dataset[k])[0] // 2:, ...] for k in keys})
            else:
                result.update({'clean/' + k: dataset[k] for k in keys})
                result.update({'noise/' + k: dataset[k] for k in keys})
        return result
